use std::env;
use std::fs;
use std::io::Read;
use clap::Parser;
use postgres::{Client, NoTls};
use base64::{engine::general_purpose::URL_SAFE_NO_PAD, Engine as _};
use serde::Serialize;
use argon2::{Argon2, PasswordHasher, password_hash::Salt};


const PASSWORD_LENGTH: usize = 64;

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    /// Setup everything from scratch
    #[arg(short, long)]
    setup: bool,
    /// Specify the directory where all the sql files are located
    #[arg(short, long)]
    psql_dir: Option<String>
}

#[derive(Serialize)]
struct Admin {
    username: String,
    password: String,
}

#[derive(Serialize)]
struct Config {
    postgres: Admin,
    bubble: Admin
}


fn random_string(size: usize) -> String {
    // Generates a URL safe string of random text of a given size
    // Internally uses /dev/urandom to generate that string
    let mut buffer = vec![0; size];
    let mut f = std::fs::File::open("/dev/urandom").unwrap();
    f.read_exact(&mut buffer).unwrap();
    URL_SAFE_NO_PAD.encode(buffer)
}

fn salt_and_hash(password: &str) -> String {
    // Generates a salted and hashed variation of the given password
    let salt_str = random_string(8);
    let salt: Salt = salt_str.as_str().try_into().unwrap();
    let a2 = Argon2::default();
    let hash = a2.hash_password(password.as_bytes(), salt).unwrap();
    hash.to_string()
}

fn admin(username: &str, password_size: usize) -> Admin {
    Admin {
        username: username.to_string(),
        password: random_string(password_size)
    }
}

fn full_setup(args: Args) -> Result<Config, postgres::Error> {
    // Check to make sure we have the DB url set to connect
    const KEY: &'static str = "DB_CONNECTION_STRING" ;
    let connection_string = env::var(KEY).expect(
        &format!("The env var {} is not set!", KEY)
    );
    let db_folder = match args.psql_dir {
        Some(val) => format!("{}/setup-tables.sql", val),
        None => "db/setup-tables.sql".to_string()
    };
    let setup_tables_script = fs::read_to_string(db_folder)
        .expect("Failed to load file: db/setup-tables.sql");
    let postgres_admin = admin(&format!("bubble_admin-{}", random_string(8)), PASSWORD_LENGTH);
    let bubble_admin = admin(
        &format!("admin-{}", random_string(8)),
        PASSWORD_LENGTH
    );

    let mut client = Client::connect(&connection_string, NoTls)?;
    // Preliminary bs
    client.execute("DROP DATABASE IF EXISTS bubble;", &[])?;
    client.execute("CREATE DATABASE bubble;", &[])?;
    client.execute("DROP USER IF EXISTS bubble_admin;", &[])?;
    let stmt = format!("CREATE ROLE bubble_admin WITH ENCRYPTED PASSWORD '{}'", postgres_admin.password);
    client.execute(&stmt , &[])?;

    // Ensure the admin has ownership of the db we created
    client.execute("ALTER DATABASE bubble OWNER TO bubble_admin", &[])?;

    // Service table creation
    client.batch_execute(&setup_tables_script)?;

    // Populate the user table with the first user ( owner )
    let salted = salt_and_hash(&bubble_admin.password);
    client.execute(
        "INSERT INTO users (id, username, password) VALUES (gen_random_uuid(), $1, $2)",
        &[&bubble_admin.username, &salted]
    )?;
    Ok(Config {
        postgres: postgres_admin,
        bubble: bubble_admin
    })
}

fn main() {
    let args = Args::parse();

    if args.setup {
        match full_setup(args) {
            Ok(config) => println!("{}", serde_json::to_string(&config).unwrap()),
            Err(e) => eprintln!("{:#?}", e)
        }
    }
}
